require 'pom/project'
require 'tmpdir'
require 'fileutils'
require 'path/shell'

module Box

  # Pack is the base class for all other packaging services.
  #
  # This is a subclass of Path::Shell to provide shell utilities.

  class Pack < Path::Shell

    attr :location

    attr :project

    attr :metadata

    # New package creator.
    #
    # Valid options:
    #
    #   :debug
    #   :noop
    #   :verbose
    #   :quiet
    #   :force
    #   :spec
    #   :manifest
    #  
    def initialize(location, options={})
      if $TRIAL || options[:trial]
        options[:noop] = true
        options[:verbose] = true
      end

      super(location, options)

      @force  = options[:force]
      @spec   = options[:spec]
      @manif  = options[:manifest]

      @project  = POM::Project.new(location, :lookup=>true) #, :load=>true)

      @location = @project.root     #dir(location)
      @metadata = @project.metadata #POM::Metadata.new(location.to_s)

      # If a dedicated package directory exists (ie. pkg/ or pack/)
      # use that instead of default locations.
      if options[:output]
        if File.directory?(options[:output])
          @pack_folder = dir(options[:output])
        else
          raise ArgumentError, "#{options[:output]} is not a directory"
        end
      elsif inside?
        # TODO: Should we go ahead and force project.pack?
        pkgdir = project.pack
        @pack_folder = dir(pkgdir) if File.directory?(pkgdir)
        #@temp_folder = @pack_folder
      else
        @pack_folder = dir(Dir.pwd) # or project.pack ?
      end
    end

    def force?    ; @force  ; end

    def spec?     ; @spec   ; end

    def manifest? ; @manif  ; end

    # TODO: Pathname should have #contains? too.
    def inside?
      FileTest.contains?(Dir.pwd, @location.to_s)
    end

    #
    def package
      unless package_needed? or force?
        report_package_already_built(package_file)
        return
      end

      if manifest? && !dryrun?
        files = metadata.distribute
        # generate manifest in root directory
        generate_manifest(location, *files)
      end

      if !manifest_file.exist?
        puts "MANIFEST file is missing"
        return
      end

      file = nil
      cd(location) do
        file = Pathname.new(package_build)
      end

      file = transfer(file, pack_folder)

      report_package_built(file)
    end

    # Overridden by subclasses for specific package formats.
    def package_build
       raise "subclass must override"
    end

    # Manifest file
    def manifest_file
      @manifest_file ||= location.glob('MANIFEST{,.txt}', :casefold).first
    end

    # Manifest list
    def manifest
      @manifest ||= (
        list = []
        manifest_file.readlines.each do |line|
          next if line =~ /^\s*$/
          next if line =~ /^\s*[#]/
          list << line.strip
        end
        list
      )
    end

    #def package_directory
    #  @package_directory ||=
    #end

    #def package_directory?
    #  package_directory.directory?
    #end

    #
    #def cache_folder
    #  #@cache_folder ||= location / '.cache'
    #  @cache_folder ||= (
    #    Pathname.new(Dir.tmpdir) / 'box' / metadata.name
    #  )
    #end

    # Unless otherwise specified temporary staging will be stored
    # in operating system temprorary folder.
    #
    def temp_folder
      @temp_folder ||= (
        Pathname.new(Dir.tmpdir) / 'box' / metadata.name
      )
    end

    # Unless otherwise specified packages will be saved to current
    # working directory.
    def pack_folder
      @pack_folder ||= location
    end

    # Current platform.
    def current_platform
      Platform.local.to_s
    end

    def extension
      raise "must override #extension"
    end

    private

    def stage_name
      metadata.package_name
    end

    def package_file
      pathname("#{stage_name}#{extension}")
    end

    # If the package needed? ie. Is it not already created?
    #
    def package_needed?
      return true unless manifest_file.exist?
      pfile = pack_folder / package_file
      files = manifest
      outofdate?(pfile, *files) # or force?
    end

    # If the package needed? ie. Is it not already created?
    #
    #def package_needed?(extension)
    #  pfile = File.join(project.pack, metadata.stage_name + extension)
    #  files = filelist
    #  outofdate?(pfile, *files) # or force?
    #end

    #
    def package_exist?(extension)
      !package_needed?(extension)
    end

    # Create staging ground.
    #
    def stage(extension, &block)
      stage_folder = temp_folder / stage_name
      #pfile = stage + extension
      #fu.rm_r(stage_folder) if File.exist?(stage_folder)   # remove old stage
      super(stage_folder.to_s, manifest)  # make new stage
      cd(stage_folder) do
        block.call if block
      end
    end

    # Transfer package file to storage location.
    #
    def transfer(file, store=nil)
      store ||= pack_folder
      # move to store, unless already there
      from = file #.path
      dest = store / file.basename
      unless from == dest
        mkdir_p(store) unless File.exist?(store)
        mv(file, store)
      end
      return dest
    end

    #
    def stage_folder
      temp_folder / stage_name
    end

    # Report that a package has been built.
    #
    # FIXME: will type, name and version always be right?
    #
    def report_package_built(file)
      type = file.extname
      file = file.basename.to_s
      name = file[0...file.rindex('-')]
      vers = file[file.rindex('-')+1..-1].chomp(type)

      puts
      puts "  Successfully built #{type}"
      puts "  Name: #{name}"
      puts "  Version: #{vers}"
      puts "  File: #{file}"
      puts
    end

    # Report that a package has been built.
    #
    # FIXME: will type, name and version always be right?
    #
    def report_package_already_built(file)
      type = file.extname
      file = file.basename.to_s
      name = file[0...file.rindex('-')]
      vers = file[file.rindex('-')+1..-1].chomp(type)

      puts
      puts "  Package #{file} is already current."
      puts
      #puts "  Name: #{name}"
      #puts "  Version: #{vers}"
      #puts "  File: #{file}"
    end

  # --- Prior to using the MANIFEST this was used.

    # Directories that a typically excluded from a distribution.
    #DEFAULT_EXCLUDE = %{ doc/log log gen temps site tmp web website work }

    # File pattern are files/dirs that are typically ignored.
    #DEFAULT_IGNORE  = %{ .* }

    # List of files and/or directories to include in package.
    #attr_accessor :distribute

    # Alias for #distribute.
    #alias_accessor :include, :distribute

    # List of files and/or directories to exclude from package.
    #attr_accessor :exclude

    # Standard files to ignore. Defaults to hidden files (.*).
    #attr_accessor :ignore

    #def initialize_defaults
    #  super
      #@distribute = ['**/*']
      #@exclude    = DEFAULT_EXCLUDE
      #@ignore     = DEFAULT_IGNORE
    #end

    # Glob of files to be distributed in package.
    # This can be a single glob or a list of globs.
    # The default is '**/*'.
    #def distribute=(val)
    #  @distribute = [val].flatten
    #end

    # Glob of files to be exclude from package.
    # This can be a single glob or a list of globs.
    # The default is to exclude the 'admin' directory.
    #def exclude=(val)
    #  @exclude = [val].flatten
    #end

    # Prepare for packaging (clean, distclean, version stamp).
    #
    # TODO: When we add support for binary packages distclean
    #       should not be done for them.

    #def prepare(options)
    #  @prepared ||= (
    #    project.clean
    #    project.make_distclean if project.compiles?
    #    project.stamp(options)
    #    true
    #  )
    #end

    ## Generate manifest.
    #def package_manifest(dir, files='**/*')
    #  return if dryrun?
    #  cd(dir) do
    #    files = Dir[files]
    #    rm('MANIFEST') if File.exist?('MANIFEST')
    #    File.open('MANIFEST', 'w') do |f|
    #      f << files.join("\n")
    #    end
    #  end
    #end

    # Generate manifest.
    #
    def generate_manifest(dir, *files)
      return if dryrun?
      files = files.flatten.compact
      files = ['**/*'] if files.empty?
      cd(dir) do
        files = multiglob_r(*files)
        rm('MANIFEST') if File.exist?('MANIFEST')
        File.open('MANIFEST', 'w') do |f|
          f << files.join("\n")
        end
      end
    end

    # List of files included in the package. This is generated using
    # ++include++ and ++exlude++.
    #
    #def filelist
    #  @filelist ||= collect_files(true)
    #end

    # Collect distribution files.
    #
    #def collect_files(with_dirs=false)
    #  files = []
    #
    #  cd(project.source) do
    #    files += Dir.multiglob_r(*distribute)
    #    files -= Dir.multiglob_r(*exclude)
    #    files -= Dir.multiglob_r(*ignore)
    #    #files -= Dir.multiglob_r(project.pack.to_s) #package_directory
    #  end
    #
    #  # Do not include symlinks.
    #  files.reject!{ |f| FileTest.symlink?(f) }
    #  # Option to exclude directories
    #  unless with_dirs
    #    files = files.select{ |f| !File.directory?(f) }
    #  end
    #  return files
    #end

    # Internal status report.
    #
    # Only output if dryrun or trace mode.
    #
    def status(message)
      if dryrun? or trace?
        puts message
      end
    end

=begin
    # Stage by hard linking included files to a stage directory.
    #
    #   stage_directory   Stage directory.
    #   source_directory  Directory in which files are located.
    #   files             Files to link to stage.
    #
    # TODO: Add options for :verbose, :noop and :dryrun ?
    #
    def stage(stage_directory, source_directory, files)
      # Ensure existance of staging area.
      fu.rm_r(stage_directory) if File.directory?(stage_directory)
      fu.mkdir_p(stage_directory)
      # Link files into staging area.
      files.each do |f|
        src  = File.join(source_directory, f)
        file = File.join(stage_directory, f)
        if File.directory?(src)
          fu.mkdir_p(file)
        else
          fu.mkdir_p(File.dirname(file))
          unless File.exist?(file) and File.mtime(file) >= File.mtime(src)
            fu.ln(src, file) #safe_ln ?
          end
        end
      end
      return stage_directory
    end
=end

  end

end

